﻿using HarmonyLib;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using UnityEngine;
using VRM;

// DONE ver1.2: 顔の向きがなんか変
// DONE ver1.2: 目パチしてくれない
// DONE ver1.2: 視線がなんか変
// DONE ver1.3: 一人称視点時のカメラ位置・表示非表示
// DONE ver1.3: リムライトとエミッシブが夜に目立つ
// DONE ver1.4: 関節の向き（IKゴール位置）を直す
// DONE ver1.6: 一人称視点で視点が振動するバグを修正
// DONE ver1.7:  足が捻じれるバグ
// TODO: 手のアニメーション同期
// TODO: 表情変化
// DONE ver1.7: 首無し影が気になる
// TODO: 足が捻じれる（多分背の高いモデル）
// TODO: 宙に浮く


class VRMData
{
	public static Dictionary<string, GameObject> VRMModelDic = new Dictionary<string, GameObject>();
	public static Dictionary<string, GameObject> InstModelDic = new Dictionary<string, GameObject>();
	public static Dictionary<Player, string> PlayerNameDic = new Dictionary<Player, string>();
	public static string LocalPlayerName;
}

static class GameState
{
	public static bool isThirdPerson = false;
}

[HarmonyPatch(typeof(Player), "Start")]
class Patch_Player_Start
{
	[HarmonyPostfix]
	static async void Postfix(Player __instance)
	{
		var networkPlayer = AccessUtil.GetField<Player, Network_Player>(__instance, "playerNetwork");
		//var playerId = networkPlayer.Network.GetIDFromPlayer(networkPlayer);
		var playerName = networkPlayer.name.Replace("PlayerStats, ", "");
		playerName = playerName.Replace("Network_Player, ", "");
		Debug.Log("[RaftVRM] playerName = " + playerName);

		var isLocalPlayer = networkPlayer.IsLocalPlayer;
		if (isLocalPlayer)
        {
			VRMData.LocalPlayerName = playerName;
        }

		var vrm = await LoadVRM(playerName);
		if (vrm != null)
		{
			var realtimeSettings = RaftVRMSettings.Init(playerName);
			VRMData.PlayerNameDic[__instance] = playerName;
			VisibleSkinnedMesh(__instance.gameObject, false);
			vrm.transform.SetParent(__instance.transform, false);
			VisibleSkinnedMesh(vrm, true);

			// 設定ファイル
			if (realtimeSettings)
            {
				vrm.AddComponent<VRMSettings>().Setup(playerName);
            }

			// アニメーション同期
			vrm.AddComponent<PlayerAnimSync>().Setup(__instance, __instance.transform.FindChildRecursively("Skeleton"), __instance.GetComponentInChildren<PlayerAnimator>().anim, vrm.GetComponentInChildren<Animator>());

			// 瞬き
			vrm.AddComponent<Blinker>();

			// マテリアルコントロール
			vrm.AddComponent<VRMMaterialController>().Setup(playerName);

            // 視線コントロール
            if (isLocalPlayer)
            {
                var offsetter = new GameObject("PositionOffsetter").AddComponent<PositionOffsetter>();
                offsetter.target = Camera.main.transform;
                offsetter.offset = Vector3.up * 0.9f + Vector3.forward * 0.1f;
                offsetter.transform.SetParent(vrm.transform, false);

                var lookAtHead = vrm.GetComponent<VRMLookAtHead>();
                if (lookAtHead == null)
                {
                    lookAtHead = vrm.AddComponent<VRMLookAtHead>();
                }
                lookAtHead.UpdateType = UpdateType.LateUpdate;
                lookAtHead.Target = offsetter.transform;
                //lookAtHead.SetEnabledSafe(false);

                var lookAtBoneApplyer = vrm.GetComponent<VRMLookAtBoneApplyer>();
                if (lookAtBoneApplyer == null)
                {
                    lookAtBoneApplyer = vrm.AddComponent<VRMLookAtBoneApplyer>();
                }
                if (lookAtBoneApplyer.LeftEye.Transform == null || lookAtBoneApplyer.RightEye.Transform == null)
                {
                    var animator = vrm.GetComponent<Animator>();
                    lookAtBoneApplyer.LeftEye = OffsetOnTransform.Create(animator.GetBoneTransform(HumanBodyBones.LeftEye));
                    lookAtBoneApplyer.RightEye = OffsetOnTransform.Create(animator.GetBoneTransform(HumanBodyBones.RightEye));
                    lookAtBoneApplyer.HorizontalOuter = new CurveMapper(90.0f, 16.0f);
                    lookAtBoneApplyer.HorizontalInner = new CurveMapper(90.0f, 16.0f);
                    lookAtBoneApplyer.VerticalDown = new CurveMapper(90.0f, 16.0f);
                    lookAtBoneApplyer.VerticalUp = new CurveMapper(90.0f, 16.0f);
                }
                //lookAtBoneApplyer.SetEnabledSafe(false);
            }

            // SpringBone調整
            vrm.AddComponent<VRMSpringBoneController>().Setup(playerName);
		}
	}

	static async Task<GameObject> LoadVRM(string playerName)
	{
		if (VRMData.VRMModelDic.ContainsKey(playerName))
		{
			var inst = UnityEngine.Object.Instantiate(VRMData.VRMModelDic[playerName]);
			VRMData.InstModelDic[playerName] = inst;
			inst.SetActive(true);
			return inst;
		}

		var path = PathUtil.Combine(RaftVRM.RaftVRMPath, playerName + ".vrm");
		Debug.Log("VRM Path = " + path);
		if (!File.Exists(path))
        {
			Debug.LogWarning("VRM file not found...");
			return null;
		}
		var loaded = await VRM.VrmUtility.LoadAsync(path);
		if (loaded == null) return null;

		loaded.ShowMeshes();
		loaded.EnableUpdateWhenOffscreen();
		var vrm = loaded.gameObject;
		UnityEngine.GameObject.DontDestroyOnLoad(vrm);
		vrm.SetActive(false);
		VRMData.VRMModelDic[playerName] = vrm;

		var res = UnityEngine.Object.Instantiate(VRMData.VRMModelDic[playerName]);
		VRMData.InstModelDic[playerName] = res;
		res.SetActive(true);
		return res;
	}

	private static readonly string[] ignoreVisibleModels = new string[]
	{
		"Notebook",
		"Pouch",
		"model",
	};

	static void VisibleSkinnedMesh(GameObject gameObject, bool visible)
	{
		var smrs = gameObject.GetComponentsInChildren<SkinnedMeshRenderer>();
		foreach (var smr in smrs)
		{
			if (!Array.Exists(ignoreVisibleModels, str => smr.name == str))
			{
				smr.enabled = visible;
			}
		}
	}

	static void LogChildren(Transform self, int depth = 0, int maxDepth = 999)
	{
		var count = self.childCount;
		for (var i = 0; i < count; i++)
		{
			var trans = self.GetChild(i);
			var sb = new StringBuilder();
			for (var d = 0; d < depth; d++) sb.Append("  ");
			sb.Append(trans.name);
			Log(sb.ToString());
			if (depth < maxDepth) LogChildren(trans, depth + 1, maxDepth);
		}
	}

	static void Log(string text)
	{
		File.AppendAllLines("D:/Steam/steamapps/common/Raft/Debug.Log", new[] { text });
	}
}

[HarmonyPatch(typeof(Player), "Update")]
class Patch_Player_Update
{
	[HarmonyPostfix]
	static void Postfix(Player __instance)
	{
		if (VRMData.PlayerNameDic.ContainsKey(__instance))
		{
            var vrmModel = VRMData.InstModelDic[VRMData.PlayerNameDic[__instance]];
			if (vrmModel == null) return;
            vrmModel.transform.position = __instance.transform.position;
        }
	}
}

[HarmonyPatch(typeof(ThirdPerson), "Update")]
class Patch_ThirdPerson_Update
{
	private static bool init = false;
	[HarmonyPostfix]
	static async void Postfix(ThirdPerson __instance)
	{
        var networkPlayer = AccessUtil.GetField<ThirdPerson, Network_Player>(__instance, "playerNetwork");
        if (networkPlayer.IsLocalPlayer && VRMData.PlayerNameDic.TryGetValue(networkPlayer.PlayerScript, out var playerName) && playerName == VRMData.LocalPlayerName)
        {
            if (!init)
            {
                init = true;
                __instance.ForceThirdPersonState(true);
                await Task.Yield();
                __instance.ForceThirdPersonState(false);
                await Task.Yield();
            }
            GameState.isThirdPerson = __instance.ThirdPersonModel;
            __instance.localCameraOffset = RaftVRMSettings.ReadVector3(VRMData.LocalPlayerName, "ThirdPersonCameraOffset", new Vector3(0.0f, 0.5f, -2.0f));
            __instance.pivotYOffset = RaftVRMSettings.ReadFloat(VRMData.LocalPlayerName, "ThirdPersonCameraLookYOffset", 0.5f);
        }
    }
}

[HarmonyPatch(typeof(CharacterModelModifications), "Update")]
class Patch_CharacterModelModifications_Update
{
	[HarmonyPostfix]
	static void Postfix(CharacterModelModifications __instance)
	{
        var networkPlayer = AccessUtil.GetField<CharacterModelModifications, Network_Player>(__instance, "playerNetwork");
        if (networkPlayer.IsLocalPlayer && VRMData.PlayerNameDic.TryGetValue(networkPlayer.PlayerScript, out var playerName) && playerName == VRMData.LocalPlayerName)
        {
            var offset = RaftVRMSettings.ReadVector3(playerName, "FirstPersonCameraOffset", Vector3.zero);
			if (offset != Vector3.zero)
            {
				__instance.cameraHolder.localPosition = new Vector3(-offset.y, offset.x, offset.z);
			}
        }
    }
}