﻿using RootMotion;
using RootMotion.FinalIK;
using System.Collections.Generic;
using System.IO;
using System.Threading.Tasks;
using UniHumanoid;
using UnityEngine;
using VRM;

[DefaultExecutionOrder(int.MaxValue)]
class ModelScaleAdjuster : MonoBehaviour
{
    private string playerName;
    private Animator orgAnim;
    private Animator vrmAnim;
    private PlayerAnimSync animSync;

    public void Setup(PlayerAnimSync animSync, string playerName, Animator vrmAnim, Animator orgAnim)
    {
        this.playerName = playerName;
        this.vrmAnim = vrmAnim;
        this.orgAnim = orgAnim;
        this.animSync = animSync;
    }

    void LateUpdate()
    {
        Adjust();
    }

    private void Adjust()
    {
        var orgScale = RaftVRMSettings.ReadFloat(playerName, "OriginalModelScale", 1.0f);
        var vrmScale = RaftVRMSettings.ReadFloat(playerName, "VRMModelScale", 1.0f);

        orgAnim.transform.localScale = Vector3.one * orgScale;
        vrmAnim.transform.localScale = Vector3.one * vrmScale;

        // つま先が捻じれる問題回避
        animSync.orgBoneDic[HumanBodyBones.LeftFoot].localRotation *= Quaternion.Inverse(Quaternion.Euler(0.0f, -120.0f, 90.0f));
        animSync.orgBoneDic[HumanBodyBones.RightFoot].localRotation *= Quaternion.Inverse(Quaternion.Euler(0.0f, -120.0f, 90.0f));
    }
}

class PlayerAnimSync : MonoBehaviour
{
    private Player player;
    private Transform orgModel;
    private Animator orgAnim;
    private Animator vrmAnim;

    public Dictionary<HumanBodyBones, Transform> orgBoneDic = new Dictionary<HumanBodyBones, Transform>();
    public Dictionary<HumanBodyBones, float> orgBoneDistRateDic = new Dictionary<HumanBodyBones, float>();
    public Dictionary<HumanBodyBones, Vector3> orgBoneInitPosDic = new Dictionary<HumanBodyBones, Vector3>();

    private FullBodyBipedIK bodyIK;
    private bool isLocalPlayer;
    private string playerName;

    public Vector3 vrmHipsLocalPosition;

    public async void Setup(Player player, Transform orgModel, Animator orgAnim, Animator vrmAnim)
    {
        orgAnim.enabled = false;
        await Task.Yield();

        this.player = player;
        this.orgModel = orgModel;
        this.orgAnim = orgAnim;
        this.vrmAnim = vrmAnim;
        this.playerName = VRMData.PlayerNameDic[player];

        orgBoneDic[HumanBodyBones.Hips] = orgModel.FindChildRecursively("ORG-hips");
        orgBoneDic[HumanBodyBones.Spine] = orgModel.FindChildRecursively("ORG-spine");
        orgBoneDic[HumanBodyBones.Chest] = orgModel.FindChildRecursively("ORG-chest");
        orgBoneDic[HumanBodyBones.Neck] = orgModel.FindChildRecursively("ORG-neck");
        orgBoneDic[HumanBodyBones.Head] = orgModel.FindChildRecursively("ORG-head");
        orgBoneDic[HumanBodyBones.LeftShoulder] = orgModel.FindChildRecursively("ORG-shoulder_L");
        orgBoneDic[HumanBodyBones.LeftUpperArm] = orgModel.FindChildRecursively("ORG-upper_arm_L");
        orgBoneDic[HumanBodyBones.LeftLowerArm] = orgModel.FindChildRecursively("ORG-forearm_L");
        orgBoneDic[HumanBodyBones.LeftHand] = orgModel.FindChildRecursively("ORG-hand_L");
        orgBoneDic[HumanBodyBones.RightShoulder] = orgModel.FindChildRecursively("ORG-shoulder_R");
        orgBoneDic[HumanBodyBones.RightUpperArm] = orgModel.FindChildRecursively("ORG-upper_arm_R");
        orgBoneDic[HumanBodyBones.RightLowerArm] = orgModel.FindChildRecursively("ORG-forearm_R");
        orgBoneDic[HumanBodyBones.RightHand] = orgModel.FindChildRecursively("ORG-hand_R");
        orgBoneDic[HumanBodyBones.LeftUpperLeg] = orgModel.FindChildRecursively("ORG-thigh_L");
        orgBoneDic[HumanBodyBones.LeftLowerLeg] = orgModel.FindChildRecursively("ORG-shin_L");
        orgBoneDic[HumanBodyBones.LeftFoot] = orgModel.FindChildRecursively("ORG-foot_L");
        orgBoneDic[HumanBodyBones.LeftToes] = orgModel.FindChildRecursively("ORG-toe_L");
        orgBoneDic[HumanBodyBones.RightUpperLeg] = orgModel.FindChildRecursively("ORG-thigh_R");
        orgBoneDic[HumanBodyBones.RightLowerLeg] = orgModel.FindChildRecursively("ORG-shin_R");
        orgBoneDic[HumanBodyBones.RightFoot] = orgModel.FindChildRecursively("ORG-foot_R");
        orgBoneDic[HumanBodyBones.RightToes] = orgModel.FindChildRecursively("ORG-toe_R");

        foreach (var kv in orgBoneDic)
        {
            //if (kv.Key == HumanBodyBones.Hips || kv.Key == HumanBodyBones.Spine) continue;

            var orgBone = kv.Value;
            var vrmBone = vrmAnim.GetBoneTransform(kv.Key);

            var orgDist = orgBone.localPosition.magnitude;
            var vrmDist = vrmBone?.localPosition.magnitude ?? orgDist;

            var distRate = Mathf.Max(vrmDist / orgDist, 0.001f);

            if (kv.Key == HumanBodyBones.LeftUpperLeg || kv.Key == HumanBodyBones.RightUpperLeg)
            {
                distRate = Mathf.Lerp(distRate, 1.0f, 0.5f);
            }

            orgBone.localPosition *= distRate;
            orgBoneDistRateDic[kv.Key] = distRate;
            orgBoneInitPosDic[kv.Key] = orgBone.localPosition;
            //Debug.Log($"{kv.Key.ToString()}: distRate={distRate}, localPos={orgBone.localPosition}, vrmLocalPos={vrmBone?.localPosition}");
        }

        vrmHipsLocalPosition = vrmAnim.GetBoneTransform(HumanBodyBones.Hips).localPosition;

        //var leftLowerArmVrmTrans = new GameObject("LeftLowerArmAdjuster").transform;
        //leftLowerArmVrmTrans.SetParent(vrmAnim.GetBoneTransform(HumanBodyBones.LeftUpperArm));
        //var rightLowerArmVrmTrans = new GameObject("RightLowerArmAdjuster").transform;
        //rightLowerArmVrmTrans.SetParent(vrmAnim.GetBoneTransform(HumanBodyBones.RightUpperArm));

        bodyIK = vrmAnim.gameObject.AddComponent<FullBodyBipedIK>();
        var references = new BipedReferences();
        BipedReferences.AutoDetectReferences(ref references, vrmAnim.transform, BipedReferences.AutoDetectParams.Default);
        bodyIK.SetReferences(references, null);
        const float positionWeight = 1.0f;
        var positionOffset = Vector3.zero;

        //bodyIK.solver.headMapping.bone = orgBoneDic[HumanBodyBones.Head];
        //bodyIK.solver.headMapping.maintainRotationWeight = 1.0f;

        bodyIK.solver.bodyEffector.target = orgBoneDic[HumanBodyBones.Spine];
        bodyIK.solver.bodyEffector.rotationWeight = 1.0f;
        bodyIK.solver.bodyEffector.positionWeight = positionWeight;
        //bodyIK.solver.bodyEffector.positionOffset = positionOffset;

        bodyIK.solver.leftHandEffector.target = orgBoneDic[HumanBodyBones.LeftHand];
        bodyIK.solver.leftHandEffector.rotationWeight = 0.0f;
        bodyIK.solver.leftHandEffector.positionWeight = positionWeight;
        //bodyIK.solver.leftHandEffector.positionOffset = positionOffset;

        bodyIK.solver.leftShoulderEffector.target = orgBoneDic[HumanBodyBones.LeftUpperArm];
        bodyIK.solver.leftShoulderEffector.rotationWeight = 0.0f;
        bodyIK.solver.leftShoulderEffector.positionWeight = positionWeight;
        //bodyIK.solver.leftShoulderEffector.positionOffset = positionOffset;

        bodyIK.solver.leftArmChain.bendConstraint.bendGoal = orgBoneDic[HumanBodyBones.LeftLowerArm];
        bodyIK.solver.leftArmChain.bendConstraint.weight = positionWeight;

        bodyIK.solver.rightHandEffector.target = orgBoneDic[HumanBodyBones.RightHand];
        bodyIK.solver.rightHandEffector.rotationWeight = 0.0f;
        bodyIK.solver.rightHandEffector.positionWeight = positionWeight;
        //bodyIK.solver.rightHandEffector.positionOffset = positionOffset;

        bodyIK.solver.rightShoulderEffector.target = orgBoneDic[HumanBodyBones.RightUpperArm];
        bodyIK.solver.rightShoulderEffector.rotationWeight = 0.0f;
        bodyIK.solver.rightShoulderEffector.positionWeight = positionWeight;
        // bodyIK.solver.rightShoulderEffector.positionOffset = positionOffset;

        bodyIK.solver.rightArmChain.bendConstraint.bendGoal = orgBoneDic[HumanBodyBones.RightLowerArm];
        bodyIK.solver.rightArmChain.bendConstraint.weight = positionWeight;

        bodyIK.solver.leftFootEffector.target = orgBoneDic[HumanBodyBones.LeftFoot];
        bodyIK.solver.leftFootEffector.rotationWeight = 1.0f;
        bodyIK.solver.leftFootEffector.positionWeight = positionWeight;
        //bodyIK.solver.leftFootEffector.positionOffset = positionOffset;

        bodyIK.solver.leftThighEffector.target = orgBoneDic[HumanBodyBones.LeftUpperLeg];
        bodyIK.solver.leftThighEffector.rotationWeight = 0.0f;
        bodyIK.solver.leftThighEffector.positionWeight = positionWeight;
        //bodyIK.solver.leftThighEffector.positionOffset = positionOffset;

        bodyIK.solver.leftLegChain.bendConstraint.bendGoal = orgBoneDic[HumanBodyBones.LeftLowerLeg];
        bodyIK.solver.leftLegChain.bendConstraint.weight = positionWeight;

        bodyIK.solver.rightFootEffector.target = orgBoneDic[HumanBodyBones.RightFoot];
        bodyIK.solver.rightFootEffector.rotationWeight = 1.0f;
        bodyIK.solver.rightFootEffector.positionWeight = positionWeight;
        //bodyIK.solver.rightFootEffector.positionOffset = positionOffset;

        bodyIK.solver.rightThighEffector.target = orgBoneDic[HumanBodyBones.RightUpperLeg];
        bodyIK.solver.rightThighEffector.rotationWeight = 0.0f;
        bodyIK.solver.rightThighEffector.positionWeight = positionWeight;
        //bodyIK.solver.rightThighEffector.positionOffset = positionOffset;

        bodyIK.solver.rightLegChain.bendConstraint.bendGoal = orgBoneDic[HumanBodyBones.RightLowerLeg];
        bodyIK.solver.rightLegChain.bendConstraint.weight = positionWeight;
        bodyIK.UpdateSolverExternal();

        var scaleAdjuster = orgModel.gameObject.AddComponent<ModelScaleAdjuster>();
        scaleAdjuster.Setup(this, playerName, vrmAnim, orgAnim);

        var handRig = vrmAnim.gameObject.AddComponent<HandRig>();
        handRig.LeftStretch = 0.0f;
        handRig.RightStretch = 0.0f;

        var networkPlayer = AccessUtil.GetField<Player, Network_Player>(player, "playerNetwork");
        isLocalPlayer = AccessUtil.GetField<Network_Player, bool>(networkPlayer, "isLocalPlayer");

        orgAnim.enabled = true;
    }

    void Update()
    {
        if (orgAnim != null && orgAnim.cullingMode != AnimatorCullingMode.AlwaysAnimate)
        {
            orgAnim.cullingMode = AnimatorCullingMode.AlwaysAnimate;
        }

        if (vrmAnim != null)
        {
            vrmAnim.transform.rotation = orgAnim.transform.rotation;
            //vrmAnim.transform.localScale = Vector3.one * RaftVRMSettings.ReadFloat(playerName, "VRMModelScale", 1.0f);

            if (isLocalPlayer)
            {
                vrmAnim.GetBoneTransform(HumanBodyBones.Head).localScale = GameState.isThirdPerson ? Vector3.one : (Vector3.one * 0.001f);
            }
        }

        foreach (var kv in orgBoneDic)
        {
            var orgBone = kv.Value;
            orgBone.localPosition = orgBoneInitPosDic[kv.Key];
        }

        // Idle: 815305476
        // Jump: 316812144
        // Rand: -878452025
    }

    private bool prevThirdPersonState = true;
    void LateUpdate()
    {
        var orgModelScale = RaftVRMSettings.ReadFloat(playerName, "OriginalModelScale", 1.0f);
        var vrmModelScale = RaftVRMSettings.ReadFloat(playerName, "VRMModelScale", 1.0f);
        var orgBoneLengthRate = RaftVRMSettings.ReadFloat(playerName, "OriginalBoneLengthRate", 0.2f);
        var hipHeightRate = RaftVRMSettings.ReadFloat(playerName, "HipHeightRate", 1.0f);
        foreach (var kv in orgBoneDic)
        {
            var orgBone = kv.Value;
            if (kv.Key == HumanBodyBones.Hips)
            {
                var hipsOffset = (vrmAnim.GetBoneTransform(HumanBodyBones.Hips).position - vrmAnim.transform.position).normalized;
                var dot = hipsOffset;
                dot.y = Mathf.Abs(dot.y);
                hipsOffset *= Mathf.Clamp01(Vector3.Dot(Vector3.up, dot));
                hipsOffset *= RaftVRMSettings.ReadFloat(playerName, "HipsYOffset");
                orgBone.position += hipsOffset;
            }

            if (vrmAnim.GetBoneTransform(HumanBodyBones.LeftToes) != null && kv.Key == HumanBodyBones.LeftFoot)
            {
                orgBone.localRotation *= Quaternion.Euler(0.0f, -120.0f, 90.0f);
            }
            if (vrmAnim.GetBoneTransform(HumanBodyBones.RightToes) != null && kv.Key == HumanBodyBones.RightFoot)
            {
                orgBone.localRotation *= Quaternion.Euler(0.0f, -120.0f, 90.0f);
            }

            float rate = (Mathf.Lerp(orgBoneDistRateDic[kv.Key], 1.0f, orgBoneLengthRate) / orgModelScale) * vrmModelScale;
            if (kv.Key == HumanBodyBones.Spine) rate *= hipHeightRate;

            orgBone.localPosition *= rate;
        }

        if (vrmAnim != null)
        {
            vrmAnim.GetBoneTransform(HumanBodyBones.Head).localRotation *= Quaternion.Euler(-10.0f, 0.0f, -5.0f);
            vrmAnim.GetBoneTransform(HumanBodyBones.LeftHand).localRotation *= Quaternion.Euler(-45.0f, 0.0f, 20.0f);
            vrmAnim.GetBoneTransform(HumanBodyBones.RightHand).localRotation *= Quaternion.Euler(-45.0f, 0.0f, 0.0f);

            if (!isLocalPlayer)
            {
                vrmAnim.GetBoneTransform(HumanBodyBones.Hips).localScale = Vector3.one;
            }
            else if (prevThirdPersonState != GameState.isThirdPerson)
            {
                foreach (var smr in vrmAnim.GetComponentsInChildren<SkinnedMeshRenderer>())
                {
                    smr.shadowCastingMode = GameState.isThirdPerson ? UnityEngine.Rendering.ShadowCastingMode.On : UnityEngine.Rendering.ShadowCastingMode.Off;
                }
                vrmAnim.GetBoneTransform(HumanBodyBones.Hips).localScale = Vector3.one * ((RaftVRMSettings.ReadBool(playerName, "FirstPersonVisibleVRM", true) || GameState.isThirdPerson) ? 1.0f : 0.001f);
            }
        }

        prevThirdPersonState = GameState.isThirdPerson;
    }
}

//ORG - hips
//  ORG - spine
//    ORG - chest
//      ORG - neck
//        ORG - head
//      ORG - shoulder_L
//        ORG - upper_arm_L
//          ORG - forearm_L
//            ORG - hand_L
//              ORG - palm_01_L
//                ORG - f_index_01_L
//                  ORG - f_index_02_L
//                    ORG - f_index_03_L
//                ORG - thumb_01_L
//                  ORG - thumb_02_L
//                    ORG - thumb_03_L
//              ORG - palm_02_L
//                ORG - f_middle_01_L
//                  ORG - f_middle_02_L
//                    ORG - f_middle_03_L
//              ORG - palm_03_L
//                ORG - f_ring_01_L
//                  ORG - f_ring_02_L
//                    ORG - f_ring_03_L
//              ORG - palm_04_L
//                ORG - f_pinky_01_L
//                  ORG - f_pinky_02_L
//                    ORG - f_pinky_03_L
//      ORG - shoulder_R
//        ORG - upper_arm_R
//          ORG - forearm_R
//            ORG - hand_R
//              ORG - palm_01_R
//                ORG - f_index_01_R
//                  ORG - f_index_02_R
//                    ORG - f_index_03_R
//                ORG - thumb_01_R
//                  ORG - thumb_02_R
//                    ORG - thumb_03_R
//              ORG - palm_02_R
//                ORG - f_middle_01_R
//                  ORG - f_middle_02_R
//                    ORG - f_middle_03_R
//              ORG - palm_03_R
//                ORG - f_ring_01_R
//                  ORG - f_ring_02_R
//                    ORG - f_ring_03_R
//              ORG - palm_04_R
//                ORG - f_pinky_01_R
//                  ORG - f_pinky_02_R
//                    ORG - f_pinky_03_R
//  ORG - thigh_L
//    ORG - shin_L
//      ORG - foot_L
//        ORG - toe_L
//  ORG - thigh_R
//    ORG - shin_R
//      ORG - foot_R
//        ORG - toe_R