﻿using HarmonyLib;
using System;
using System.IO;
using System.Linq;
using System.Reflection;
using UnityEngine;

public class RaftVRM : Mod
{
    private Harmony harmonyInstance;

    public static string RaftVRMPath => PathUtil.Combine(Path.GetDirectoryName(Application.dataPath), "mods", "RaftVRMData");
    public static string RaftManagedPath => PathUtil.Combine(Path.GetDirectoryName(Application.dataPath), "Raft_Data", "Managed");

    private static readonly string[] PluginNames = new string[]
    {
        "FastSpringBone.dll",
        "MToon.dll",
        "netstandard.dll",
        "RootMotion.dll",
        "UniGLTF.dll",
        "UniHumanoid.dll",
        "VRM.dll",
        "VRMShaders.GLTF.IO.Runtime.dll",
        "VRMShaders.GLTF.UniUnlit.Runtime.dll",
        "VRMShaders.VRM.IO.Runtime.dll",
        "VRMShaders.VRM10.Format.Runtime.dll",
        "VRMShaders.VRM10.MToon10.Runtime.dll",
    };

    public void Start()
    {
        // RaftVRMDataフォルダを作る
        if (!Directory.Exists(RaftVRMPath))
        {
            Directory.CreateDirectory(RaftVRMPath);
        }

        // template設定ファイルを作成&更新
        try
        {
            var settingsTxt = GetEmbeddedFileBytes("settings_template.txt");
            File.WriteAllBytes(PathUtil.Combine(RaftVRMPath, "settings_template.txt"), settingsTxt);
        }
        catch
        { }

        // Pluginをインストール
        var installOnly = false;
        foreach (var pluginName in PluginNames)
        {
            var path = PathUtil.Combine(RaftManagedPath, pluginName);
            var buf = GetEmbeddedFileBytes(pluginName);
            if (!File.Exists(path) || !buf.SequenceEqual(File.ReadAllBytes(path)))
            {
                installOnly = true;
                Debug.Log("[RaftVRM] Install plugin: " + pluginName);
                File.WriteAllBytes(path, buf);
            }
        }

        if (!installOnly)
        {
            // シェーダの読み込み
            VRMShader.Initialize(GetEmbeddedFileBytes("vrmshaders.assets"));

            harmonyInstance = new Harmony("com.yoship1639.RaftVRM");
            harmonyInstance.PatchAll(Assembly.GetExecutingAssembly());
        }
        else
        {
            Debug.Log("[RaftVRM] Please restart game or reload RaftVRM mod.");
            UnloadMod();
        }
    }

    public override void UnloadMod()
    {
        base.UnloadMod();

        VRMShader.Destroy();
        harmonyInstance?.UnpatchAll(harmonyInstance.Id);
        Destroy(gameObject);
    }
}