﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using UnityEngine;

class VRMMaterialSetting
{
    public Color rimColor;
    public Color emissiveColor;
}

class VRMMaterialController : MonoBehaviour
{
    private string playerName;

    private Dictionary<Material, VRMMaterialSetting> materialDic = new Dictionary<Material, VRMMaterialSetting>();

    public void Setup(string playerName)
    {
        this.playerName = playerName;

        foreach (var smr in GetComponentsInChildren<SkinnedMeshRenderer>(true))
        {
            foreach (var mat in smr.materials)
            {
                if (!materialDic.ContainsKey(mat))
                {
                    var settings = new VRMMaterialSetting();
                    if (mat.HasProperty("_RimColor")) settings.rimColor = mat.GetColor("_RimColor");
                    if (mat.HasProperty("_EmissionColor")) settings.emissiveColor = mat.GetColor("_EmissionColor");
                    materialDic[mat] = settings;
                }
            }
        }
    }

    private 

    void Update()
    {
        foreach (var kv in materialDic)
        {
            var mat = kv.Key;
            if (mat.HasProperty("_RimColor"))
            {
                var rim = kv.Value.rimColor * RaftVRMSettings.ReadFloat(playerName, "RimColorIntensity", 1.0f);
                mat.SetColor("_RimColor", rim);
            }
            if (mat.HasProperty("_EmissionColor"))
            {
                var emissive = kv.Value.emissiveColor * RaftVRMSettings.ReadFloat(playerName, "EmissiveColorIntensity", 1.0f);
                mat.SetColor("_EmissionColor", emissive);
            }
        }
    }
}